@@ -54,6 +54,7 @@ enum {
 	LOGD_WIMAX      = 0x00800000,
 	LOGD_INFINIBAND = 0x01000000,
 	LOGD_FIREWALL   = 0x02000000,
+	LOGD_ADSL       = 0x04000000,
 };
 
 #define LOGD_DHCP (LOGD_DHCP4 | LOGD_DHCP6)
@@ -67,8 +68,8 @@ enum {
 };
 
 typedef enum {
-    NM_LOGGING_ERROR_UNKNOWN_LEVEL = 0,
-    NM_LOGGING_ERROR_UNKNOWN_DOMAIN = 1,
+	NM_LOGGING_ERROR_UNKNOWN_LEVEL = 0,  /*< nick=UnknownLevel >*/
+	NM_LOGGING_ERROR_UNKNOWN_DOMAIN = 1, /*< nick=UnknownDomain >*/
 } NMLoggingError;
 
 #define NM_LOGGING_ERROR (nm_logging_error_quark ())
