@@ -104,6 +104,7 @@ typedef struct {
 enum {
 	PROPERTIES_CHANGED,
 	VPN_STATE_CHANGED,
+	INTERNAL_STATE_CHANGED,
 
 	LAST_SIGNAL
 };
@@ -223,7 +224,8 @@ nm_vpn_connection_set_vpn_state (NMVPNConnection *connection,
 	 */
 	g_object_ref (connection);
 
-	g_signal_emit (connection, signals[VPN_STATE_CHANGED], 0, vpn_state, old_vpn_state, reason);
+	g_signal_emit (connection, signals[VPN_STATE_CHANGED], 0, vpn_state, reason);
+	g_signal_emit (connection, signals[INTERNAL_STATE_CHANGED], 0, vpn_state, old_vpn_state, reason);
 	g_object_notify (G_OBJECT (connection), NM_VPN_CONNECTION_VPN_STATE);
 
 	switch (vpn_state) {
@@ -1600,7 +1602,15 @@ nm_vpn_connection_class_init (NMVPNConnectionClass *connection_class)
 
 	/* signals */
 	signals[VPN_STATE_CHANGED] =
-		g_signal_new (NM_VPN_CONNECTION_VPN_STATE_CHANGED,
+		g_signal_new ("vpn-state-changed",
+		              G_OBJECT_CLASS_TYPE (object_class),
+		              G_SIGNAL_RUN_FIRST,
+		              0, NULL, NULL,
+		              _nm_marshal_VOID__UINT_UINT,
+		              G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_UINT);
+
+	signals[INTERNAL_STATE_CHANGED] =
+		g_signal_new (NM_VPN_CONNECTION_INTERNAL_STATE_CHANGED,
 		              G_OBJECT_CLASS_TYPE (object_class),
 		              G_SIGNAL_RUN_FIRST,
 		              0, NULL, NULL,
